# **********************************************************************
# Copyright 2014 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import nsx_logging
import os
import json
import uuid

import cli.commands.host_shared.ovs_util as ovs_util

from cli.commands.cli_command_provider import CliCommandProvider


lg = nsx_logging.getLogger("cli.commands.host_shared.logical_switch")

def vni_label_to_id(vni_label):
    """Converts a user input vni/vlan label to logical switch ID.

    Args:
        vni_label: Logical switch vni/vlan label in the format of
                    <vni-number> or vlan:<vlan-number>

    Returns:
        Logical switch ID in the tuple format of
        (is_vni, vni/vlan number). None if input is not valid.
    """
    lower_id = vni_label.lower()
    if lower_id.startswith("vlan:"):
        lower_id = lower_id[5:]
        is_vni = False
    else:
        is_vni = True
    try:
        vni = int(lower_id)
    except ValueError:
        return None
    if vni < 0:
        return None
    return (is_vni, vni)

def resolve_ls_id(id):
    try:
        uuid_val = uuid.UUID(id)
        return ovs_util.ls_uuid_to_id(id)
    except ValueError:
        ls_id = vni_label_to_id(id)
        if ls_id is not None and ovs_util.id_exists(ls_id):
            return ls_id
        return None

class LogicalSwitchIdArgument(object):
    @staticmethod
    def verify(id_str, req_dict):
        if vni_label_to_id(id_str):
            return True
        try:
            uuid_val = uuid.UUID(id_str)
        except ValueError:
            return False
        return True

class HostGetLogicalSwitch(CliCommandProvider):
    def __call__(self, cmd_list):
        res = {}
        id = resolve_ls_id(cmd_list[2])
        if id is None:
            return 'MSG_LS_NOT_FOUND'
        res['id'] = id[1]
        res['is_vni'] = id[0]
        res['uuid'] = ovs_util.ls_id_to_uuid(id)
        res['mode'] = ovs_util.get_replication_mode(id[1])
        res['name'] = ''
        ls_list = ovs_util.get_lswitch_list()
        for lswitch in ls_list:
            ls_uuid = lswitch[0]
            if res['uuid'] == ls_uuid:
                res['name'] = lswitch[3]
                break
        if id[0]:
            vni_entry = ovs_util.table_to_map(
                ovs_util.get_vni_list(), 0)[str(id[1])][0]
        else:
            vni_entry = ovs_util.table_to_map(
                ovs_util.get_vlan_list(), 0)[str(id[1])][0]
        assert(int(vni_entry[0]) == id[1])
        res['controller_ip'] = vni_entry[1]
        res['link_status'] = vni_entry[2]
        res['num_vifs'] = len(ovs_util.get_lport_list(id))
        if id[0]:
            res['num_macs'] = len(ovs_util.get_mac_table(id[1]))
            res['num_teps'] = len(ovs_util.get_vtep_list(id[1]))
            res['num_arps'] = len(ovs_util.get_arp_table(id[1]))
        return "OK\n\n%s" % json.dumps(res)

class HostGetLogicalSwitchPortTable(CliCommandProvider):
    def __call__(self, cmd_list):
        res = []
        ls_id = resolve_ls_id(cmd_list[2])
        if ls_id is None:
            return 'MSG_LS_NOT_FOUND'
        lp_list = ovs_util.get_lport_list(ls_id)
        for entry in lp_list:
            lp_uuid = entry[0]
            status = ovs_util.get_lport_status(lp_uuid)
            vif = ovs_util.get_lport_vif(lp_uuid)
            snoops = ovs_util.get_lport_snoop_mode(lp_uuid)
            modes = []
            if snoops[0]:
                modes.append("arp");
            if snoops[1]:
                modes.append("dhcp");
            if snoops[2]:
                modes.append("spoof-guard");
            if snoops[3]:
                modes.append("nd_snooping")
            if snoops[4]:
                modes.append("dhcpv6_snooping")
            res.append({'uuid': lp_uuid, 'status': status, 'vif': vif,
                        'snoop': ', '.join(modes)})
        return "OK\n\n%s" % json.dumps(res)

class HostGetLogicalSwitchMACTable(CliCommandProvider):
    def __call__(self, cmd_list):
        res = []
        ls_id = resolve_ls_id(cmd_list[2])
        if ls_id is None:
            return 'MSG_LS_NOT_FOUND'
        elif not ls_id[0]:
            return 'MSG_VLAN_NOT_SUPPORTED'
        mac_table = ovs_util.get_mac_table(ls_id[1])
        for entry in mac_table:
            res.append({'mac':entry[0], 'vtep':entry[1]})
        return "OK\n\n%s" % json.dumps(res)

class HostGetLogicalSwitchVTEPs(CliCommandProvider):
    def __call__(self, cmd_list):
        res = []
        ls_id = resolve_ls_id(cmd_list[2])
        if ls_id is None:
            return 'MSG_LS_NOT_FOUND'
        elif not ls_id[0]:
            return 'MSG_VLAN_NOT_SUPPORTED'
        vtep_table = ovs_util.get_vtep_table(ls_id[1])
        for entry in vtep_table:
            res.append({'label':int(entry[0]), 'ip':entry[1], 'mac':entry[2]})
        return "OK\n\n%s" % json.dumps(res)

class HostGetLogicalSwitchARPTable(CliCommandProvider):
    def __call__(self, cmd_list):
        res = []
        ls_id = resolve_ls_id(cmd_list[2])
        if ls_id is None:
            return 'MSG_LS_NOT_FOUND'
        elif not ls_id[0]:
            return 'MSG_VLAN_NOT_SUPPORTED'
        arp_table = ovs_util.get_arp_table(ls_id[1])
        for entry in arp_table:
            res.append({'ip':entry[0], 'mac':entry[1]})
        return "OK\n\n%s" % json.dumps(res)
