# ***************************************************************************
# Copyright 2014-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************


# This module originally developed in NSXmh cloudnet repo:
# cloudnet/src/python/management_api/descriptors/module_descriptor.py


__all__ = ["ModuleDescriptor", "Module"]


import nsx_logging
import os
import sys
import yaml

from cli.descriptors.descriptor import Descriptor
from cli.descriptors.error import DescriptorException
from cli.descriptors.type_descriptor import TypeDescriptor
from cli.utils.constants import PICKLE_DUMPS_DIR as _PICKLE_DUMPS_DIR
from cli.utils.constants import PICKLE_EXTENSION as _PICKLE_EXTENSION
from cli.utils.node_utils import get_node_type
from compat23.compat23_utils import compat23_fopen
from compat23.compat23_utils import compat23_string_types
from compat23.compat23_utils import compat23_pickle
from cli.utils.error_code import ErrorCode as ec


lg = nsx_logging.getLogger("cli.descriptors.module_descriptor")


ANY_VERSION = "_any_"

MODULE_SCHEMA_FILE = os.path.join(os.path.dirname(__file__),
                                  'module_descriptor.yml')
MODULE_SCHEMA = yaml.load(open(MODULE_SCHEMA_FILE), Loader=yaml.Loader)


class ModuleDescriptor(Descriptor):
    """Module meta information."""

    yaml_tag = '!Module'

    spec_schema = MODULE_SCHEMA

    @classmethod
    def from_raw_spec(cls, raw_spec):
        params = dict([(k, v) for k, v in list(raw_spec.items()) if k != 'id'])
        return cls(raw_spec['id'], **params)

    def __init__(self, id, description=None, revisions=None,
                 organization=None, contact=None, min_version=ANY_VERSION,
                 max_version=ANY_VERSION, snmp_oid=None, nsx_feature=None):
        self.id = id
        self.description = description
        self.revisions = revisions
        self.organization = organization
        self.contact = contact
        self.min_version = str(min_version)
        self.max_version = str(max_version)
        self.snmp_oid = snmp_oid
        self._nsx_feature = nsx_feature


class Module(object):
    """Bundle of interface descriptors with similar purpose and scope.
    """

    def __init__(self, module_file, descriptor_factory,
                 load_for_pickle_build=False, **kwargs):
        """Initialize module from raw definition."""
        # If we are being invoked during the build process to generate
        # a pickle file of descriptors included in module_file, no other
        # initialization is needed, exit now
        self._NODE_TYPE = get_node_type()
        self._pickle_dumps = {}
        self._load_for_pickle_build = load_for_pickle_build
        if load_for_pickle_build:
            return

        if isinstance(module_file, compat23_string_types):
            module_file = compat23_fopen(module_file, 'r', encoding="UTF-8")

        self.descriptors = {}
        self.meta = None
        module_min_version = None
        module_max_version = None

        current_module_disabled = False

        # Pickle is built for Python 3
        # Load from respective module pickle file, if available in pickle path.
        if sys.version_info.major >= 3 or kwargs.get("overwrite_version") == 3:
            file_name = os.path.split(module_file.name)[-1]
            pickle_file_name = "%s%s" % (file_name, _PICKLE_EXTENSION)
            pickle_path = os.path.join(_PICKLE_DUMPS_DIR, pickle_file_name)
            self._load_pickle_dump(pickle_path)

        for descriptor in self.load_descriptors(module_file):
            if isinstance(descriptor, ModuleDescriptor) and self.meta is None:
                self.meta = descriptor
                module_min_version = descriptor.min_version
                module_max_version = descriptor.max_version
            elif descriptor.id in self.descriptors:
                # each module should have unique descriptor names
                # uniqueness across modules (in imports) can be
                # enforced using the 'prefix' import option
                err = ("Descriptor '%s' in module '%s' declared multiple "
                       "times" % (descriptor.id, module_file.name))
                raise DescriptorException(err)
            else:
                if isinstance(descriptor, ModuleDescriptor):
                    if descriptor_factory.feature_disabled(descriptor):
                        current_module_disabled = True
                        continue
                    current_module_disabled = False
                    module_min_version = descriptor.min_version
                    module_max_version = descriptor.max_version
                if not current_module_disabled:
                    self.descriptors[descriptor.id] = descriptor

        if self.meta is None:
            self.meta = ModuleDescriptor(id=module_file.name)
        self.id = self.meta.id

    @classmethod
    def _load_pickle_file(cls, pickle_file):
        """Loads and returns data from target pickle file"""
        data = None
        try:
            with open(pickle_file, 'rb') as fd:
                data = compat23_pickle.load(fd)
        except UnicodeDecodeError:
            with open(pickle_file, 'rb') as fd:
                data = compat23_pickle.load(fd, encoding='latin1')
        except Exception:
            lg.exception("Error loading pickle dump file: %s", pickle_file,
                         error_code=ec.INTERNAL_ERROR)
        return data

    def _load_pickle_dump(self, pickle_path):
        """Adds pickle data from pickle file into module"""
        pickle_key = os.path.split(pickle_path)[-1]
        if os.path.isfile(pickle_path):
            self._add_pickle_dump(
                {pickle_key: self._load_pickle_file(pickle_path)})

    def _add_pickle_dump(self, pickle_data):
        """Updates additional pickle data into module"""
        if pickle_data:
            self._pickle_dumps.update(pickle_data)

    def load_descriptors(self, module_file):
        ret = []
        # If a pickle dump exists for the specified module_file, use it
        base_path, file_name = os.path.split(module_file.name)
        pickle_dump_key = "%s%s" % (file_name, _PICKLE_EXTENSION)
        if pickle_dump_key in self._pickle_dumps:
            return self._pickle_dumps[pickle_dump_key]

        for mlist in yaml.load_all(module_file, Loader=yaml.Loader):
            # doc may contain a single descriptor, or a list of descriptors
            if not isinstance(mlist, list):
                mlist = [mlist]
            ret.extend(Descriptor.evaluate_markup(mlist, False, base_path))

        if self._load_for_pickle_build:
            return ret

        for descriptor in ret:
            lg.debug("loaded descriptor %s from module %s"
                     % (descriptor.id, module_file.name))
        lg.debug("module %s parsed succesfully. %d descriptors loaded"
                 % (module_file.name, len(ret)))

        return ret
