# Copyright (C) 2016 VMware, Inc. All rights reserved.

import sys
import logging
import uuid

import gevent
import gevent.queue

from google.protobuf.service import RpcChannel

from vmware.nsx.rpc import nsx_rpc_pb2
from vmware.nsx.rpc.status import CallStatus
from vmware.nsx.rpc.status import CloseReason

LOGGER = logging.getLogger("nsxrpc")

SYNCHRONOUS = 1
ASYNCHRONOUS = 2


class NsxRpcCallError(Exception):
    """Raise when StatusCode is not SUCCESS"""
    def __init__(self, status):
        super(NsxRpcCallError, self).__init__("RPC Call Failed with %s" % status)
        self.status = status

    @staticmethod
    def from_code(status_code):
        return NsxRpcCallError(CallStatus(status_code, None))

    @property
    def code(self):
        return self.status.code

    @property
    def name(self):
        return self.status.name

    @property
    def message(self):
        return self.status.message


class Writer(object):

    def __init__(self, channel, call):
        LOGGER.debug("Writer::Constructor")
        self.call = call
        self.channel = channel
        self.is_first = True

    def next(self, request):
        LOGGER.debug("Writer::Next")
        assert isinstance(
            request,
            self.call.method_descriptor.input_type._concrete_class
        ), (
            "expected request type '%s' given '%s'" % (
                self.call.method_descriptor.input_type._concrete_class,
                request.__class__
            )
        )
        assert self.call.local_done is False

        if self.is_first:
            stream_control, frame, payload = self.channel.BuildFrameMsg(
                self.call.method_descriptor,
                self.call.call_id,
                request,
                client_context=self.call.client_context,
                status=None
            )

            self.is_first = False
        else:
            stream_control, frame, payload = self.channel.BuildFrameMsg(
                None,
                self.call.call_id,
                request,
                client_context=None,
                status=None
            )

        self.channel.connection.Send(self.channel.stream_id,
                                     stream_control, frame, payload)

    def done(self, request=None):
        LOGGER.debug("Writer::Done")
        assert request is None or isinstance(
            request,
            self.call.method_descriptor.input_type._concrete_class), (
                "expected request type '%s' given '%s'" % (
                    self.call.method_descriptor.input_type._concrete_class,
                    request.__class__
                )
            )
        assert self.call.local_done is False

        self.call.local_done = True

        if self.is_first:
            stream_control, frame, payload = self.channel.BuildFrameMsg(
                self.call.method_descriptor,
                self.call.call_id,
                request,
                client_context=self.call.client_context,
                status=CallStatus(CallStatus.SUCCESS, None)
            )
            self.is_first = False
        else:
            stream_control, frame, payload = self.channel.BuildFrameMsg(
                None,
                self.call.call_id,
                request,
                client_context=None,
                status=CallStatus(CallStatus.SUCCESS, None)
            )

        self.channel.connection.Send(
            self.channel.stream_id, stream_control, frame, payload
        )


class ResponseGenerator(object):

    def __init__(self, context):
        LOGGER.debug("ResponseGenerator::Constructor (call:%s)", context)
        self.context = context
        self.done = False

    def __iter__(self):
        return self

    def __next__(self, timeout=None):
        LOGGER.debug("ResponseGenerator::next")

        if self.done:
            LOGGER.warning("ResponseGenerator::next iteration is already done")
            raise StopIteration

        if timeout is None:
            timeout = self.context.timeout

        try:
            rsp = self.context.queue.get(block=True, timeout=timeout)
        except gevent.queue.Empty:
            self.context.status = CallStatus(CallStatus.TIMEOUT, None)

        if (self.context.status is not None and
                not self.context.status.is_success()):
            raise NsxRpcCallError(self.context.status)

        if rsp is None:
            LOGGER.debug("ResponseGenerator::next rsp is None")
            self.done = True
            raise StopIteration
        else:
            LOGGER.debug("ResponseGenerator::next %s", rsp.__class__.__name__)
            return rsp

    next = __next__  # Python 2


class Call(object):

    def __init__(self, channel, context):
        LOGGER.debug("Call::Constructor")
        self.context = context
        self.channel = channel
        self.responseList = ResponseGenerator(context)

    def Cancel(self):
        LOGGER.debug("Call::Cancel")
        self.channel.CancelCall(self.context.id)

    def Done(self):
        LOGGER.debug("Call::Done")
        return self.context.local_done and self.context.remote_done

    def LocalDone(self):
        LOGGER.debug("Call::LocalDone")
        return self.context.local_done

    def RemoteDone(self):
        LOGGER.debug("Call::RemoteDone")
        return self.context.remote_done

    def Status(self):
        LOGGER.debug("Call::Status")
        return self.context.status


class ClientContext(object):

    def __init__(self):
        self.client_id = None
        self.data = {}

    def SetClientId(self, client_id):
        self.client_id = client_id

    def Add(self, key, value):
        self.data[key] = value

    def Items(self):
        return list(self.data.items())


class SyncCallContext(object):
    def __init__(self, timeout, client_context):
        self.timeout = timeout
        self.client_context = client_context


class AsyncCallContext(object):
    def __init__(self, callback, client_context):
        self.callback = callback
        self.client_context = client_context


class SecurityContext(object):
    def __init__(self, certfile=None, keyfile=None, ca_certs=None):
        self.certfile = certfile
        self.keyfile = keyfile
        self.ca_certs = ca_certs


class NsxRpcChannel(RpcChannel):

    class _Call(object):
        def __init__(
                self,
                call_id=None,
                call_type=None,
                method_descriptor=None,
                response_class=None,
                callback=None,
                timeout=None,
                queue=None,
                local_done=None,
                remote_done=None,
                client_context=None,
                client_context_sent=None,
                status=None):
            self.call_id = call_id
            self.call_type = call_type
            self.method_descriptor = method_descriptor
            self.response_class = response_class
            self.callback = callback
            self.timeout = timeout
            self.queue = queue
            self.local_done = local_done
            self.remote_done = remote_done
            self.client_context = client_context
            self.client_context_sent = client_context_sent
            self.status = status

        def __str__(self):
            res = ""
            res += "call_id:%s " % self.call_id
            res += "type:%s " % self.call_type
            res += "method_descriptor:%s " % self.method_descriptor
            res += "response_class:%s " % self.response_class
            res += "callback:%s " % self.callback
            res += "timeout:%s " % self.timeout
            res += "queue:%s " % self.queue
            res += "local_done:%s " % self.local_done
            res += "remote_done:%s " % self.remote_done
            res += "client_context:%s" % self.client_context
            res += "client_context_sent:%s " % self.client_context_sent
            res += "status:%s" % self.status
            return res

    class Stats(object):

        def __init__(self):
            self.unexpected_messages_received = 0

    def __init__(self, connection, rpc_prefix, stream_destination=None):
        LOGGER.debug(
            "NsxRpcChannel::Constructor (connection:%s rpc_prefix:%s)",
            connection,
            rpc_prefix,
        )
        self.connection = connection
        self.rpc_prefix = rpc_prefix
        self.stream_destination = stream_destination
        self.call = {}
        self.stream_id = uuid.uuid4()
        self.stats = self.Stats()
        self.priority = 64
        self.next_call_id = 0
        self.Open()
        self.is_open = True

    def Parse(self, stream_id, stream_control_bytes,
              frame_bytes, payload_bytes):
        LOGGER.debug("NsxRpcChannel::Parse")

        if stream_id != self.stream_id:
            LOGGER.warning(
                "NsxRpcChannel::Parse stream id mismatch %s %s - dropped",
                stream_id,
                self.stream_id,
            )
            return

        stream_control = nsx_rpc_pb2.StreamControlMsg()
        stream_control.ParseFromString(stream_control_bytes)
        if stream_control.HasField("ack_stream") and stream_control.ack_stream:
            LOGGER.debug(
                "NsxRpcChannel::Parse stream %s opened successfully", stream_id
            )
            # TODO: do we need this logic? stream is ready to accept calls now
            # self.AckStreamEvent.set()
            return
        if (stream_control.HasField("close_stream") and
                stream_control.close_stream):
            LOGGER.info("NsxRpcChannel::Parse stream %s closed by server",
                        stream_id)
            self.Close(CloseReason(CloseReason.CLOSED_BY_REMOTE))
            return

        frame = nsx_rpc_pb2.FrameMsg()
        frame.ParseFromString(frame_bytes)

        LOGGER.debug(
            "NsxRpcChannel::Parse incoming stream_control %s frame %s",
            stream_control,
            frame
        )

        call_id = frame.rpc_msg.call_id
        if call_id not in self.call:
            LOGGER.error("call_id not found (%s)", call_id)
            self.stats.unexpected_messages_received += 1
            return

        call = self.call[call_id]

        if frame.rpc_msg.HasField('status'):
            call_status = CallStatus(
                frame.rpc_msg.status.code,
                frame.rpc_msg.status.error_msg)

            LOGGER.debug("NsxRpcChannel::Parse call %s done with status %s",
                         call_id, call_status)
            call.remote_done = True
            call.status = call_status

        payload = None
        if frame.rpc_msg.HasField('has_payload') and frame.rpc_msg.has_payload:
            payload = call.response_class()
            payload.ParseFromString(payload_bytes)
            LOGGER.debug("NsxRpcChannel::Parse payload %s", payload)

        if call.call_type is SYNCHRONOUS:
            if payload:
                call.queue.put(payload)
            if call.remote_done:
                call.queue.put(None)
                self.RemoveCallContext(call_id)
        elif call.call_type is ASYNCHRONOUS:
            if payload:
                call.callback(message=payload, status=call.status)
            if call.remote_done:
                call.callback(message=None, status=call.status)
                self.RemoveCallContext(call_id)
        else:
            LOGGER.error("NsxRpcChannel::Parse wrong call type (%s)",
                         call.call_type)
            assert False, "Unexpected call type %s" % call.call_type

    def RemoveCallContext(self, call_id):
        LOGGER.debug("NsxRpcChannel::RemoveCallContext")
        if call_id in self.call:
            del self.call[call_id]
        else:
            LOGGER.error("NsxRpcChannel::RemoveCallContext call not found (%s)",
                         call_id)

    def CancelCall(self, call_id):
        self.connection.Send(
            self.stream_id, self.BuildCancelCallFrameMsg(call_id))

    def GetClientStreaming(self, method_descriptor):
        options = method_descriptor.GetOptions()
        extension = options._extensions_by_name.get(
            'vmware.nsx.rpc.client_streaming')
        return (extension and options.HasExtension(extension)
                and options.Extensions[extension])

    def GetServerStreaming(self, method_descriptor):
        options = method_descriptor.GetOptions()
        extension = options._extensions_by_name.get(
            'vmware.nsx.rpc.server_streaming')
        return (extension and options.HasExtension(extension)
                and options.Extensions[extension])

    def CallMethod(
            self,
            method_descriptor,
            rpc_controller,
            request,
            response_class,
            context):
        LOGGER.debug("NsxRpcChannel::CallMethod")

        if self.is_open is not True:
            LOGGER.error("Stream is not open")
            raise NsxRpcCallError(CallStatus(CallStatus.COMMUNICATION_ERROR, None))

        client_streaming = self.GetClientStreaming(method_descriptor)
        server_streaming = self.GetServerStreaming(method_descriptor)

        LOGGER.debug(
            "NsxRpcChannel::CallMethod client_streaming:%s server_streaming:%s",
            client_streaming,
            server_streaming
        )

        if isinstance(context, SyncCallContext):
            LOGGER.debug(
                "NsxRpcChannel::CallMethod synchronous call (timeout:%s)",
                context.timeout)
            if server_streaming and client_streaming:
                return self.SyncCallMethodBidirectionalStreaming(
                    method_descriptor=method_descriptor,
                    # request=request,
                    response_class=response_class,
                    context=context,
                )
            elif server_streaming:
                return self.SyncCallMethodServerStreaming(
                    method_descriptor=method_descriptor,
                    request=request,
                    response_class=response_class,
                    context=context,
                )
            elif client_streaming:
                return self.SyncCallMethodClientStreaming(
                    method_descriptor=method_descriptor,
                    # request=request,
                    response_class=response_class,
                    context=context,
                )
            else:
                return self.SyncCallMethodSimple(
                    method_descriptor=method_descriptor,
                    request=request,
                    response_class=response_class,
                    context=context,
                )
        elif isinstance(context, AsyncCallContext):
            LOGGER.debug(
                "NsxRpcChannel::CallMethod asynchronous call (callback=%s)",
                context.callback)
            if server_streaming and client_streaming:
                return self.AsyncCallMethodBidirectionalStreaming(
                    method_descriptor=method_descriptor,
                    # request=request,
                    response_class=response_class,
                    context=context,
                    )
            elif server_streaming:
                return self.AsyncCallMethodServerStreaming(
                    method_descriptor=method_descriptor,
                    request=request,
                    response_class=response_class,
                    context=context,
                )
            elif client_streaming:
                return self.AsyncCallMethodClientStreaming(
                    method_descriptor=method_descriptor,
                    # request=request,
                    response_class=response_class,
                    context=context,
                )
            else:
                return self.AsyncCallMethodSimple(
                    method_descriptor,
                    request,
                    response_class,
                    context
                )
        else:
            # should never be here
            assert False, (
                "context must be either "
                "SyncCallContext or AsyncCallContext (given %s)"
                % context)

    def BuildFrameMsg(
            self,
            method_descriptor,
            call_id,
            request,
            client_context=None,
            status=None):
        LOGGER.debug("NsxRpcChannel::BuildFrameMsg")
        frame = nsx_rpc_pb2.FrameMsg()
        frame.rpc_msg.call_id = call_id

        if status is not None:
            frame.rpc_msg.status.code = status.code

        if method_descriptor is not None:
            frame.rpc_msg.method = self.rpc_prefix + '/' + method_descriptor.name

        if client_context is not None:
            LOGGER.debug("NsxRpcChannel::BuildFrameMsg client context: %s",
                         client_context)
            call = self.call[call_id]
            if call.client_context_sent:
                LOGGER.debug(
                    "NsxRpcChannel::BuildFrameMsg client context is already sent")
            else:
                LOGGER.debug("NsxRpcChannel::BuildFrameMsg populating client context")
                for key, value in client_context.Items():
                    LOGGER.debug("NsxRpcChannel::BuildFrameMsg adding pair %s:%s",
                                 key, value)
                    header = frame.header.header.add()
                    header.key = key
                    header.value = value
                if client_context.client_id:
                    frame.header.client_id = client_context.client_id
                call.client_context_sent = True

        # TODO: support empty payload
        has_payload = request is not None
        payload = None
        if has_payload:
            frame.rpc_msg.has_payload = has_payload
            payload = request.SerializeToString()

        LOGGER.debug("NsxRpcChannel::BuildFrameMsg outcoming frame %s payload %s",
                     frame, payload)

        return None, frame.SerializeToString(), payload

    def BuildCancelCallFrameMsg(self, call_id):
        frame = nsx_rpc_pb2.FrameMsg()
        frame.rpc_msg.call_id = call_id
        frame.rpc_msg.cancel_call = True
        return None, frame.SerializeToString(), None

    def NextCallId(self):
        call_id = self.next_call_id
        self.next_call_id += 1
        return call_id

    def SyncCallMethodSimple(
            self,
            method_descriptor,
            request,
            response_class,
            context):
        LOGGER.debug("NsxRpcChannel::SyncCallMethodSimple")
        call_id = self.NextCallId()

        call = self.call[call_id] = self._Call(
            call_id=call_id,
            call_type=SYNCHRONOUS,
            method_descriptor=method_descriptor,
            response_class=response_class,
            timeout=context.timeout,
            callback=None,
            queue=gevent.queue.Queue(),
            local_done=False,
            remote_done=False,
            client_context=context.client_context,
            client_context_sent=False,
            status=None,
        )

        stream_control, frame, payload = self.BuildFrameMsg(
            method_descriptor,
            call_id,
            request,
            client_context=call.client_context,
            status=None
        )

        self.connection.Send(self.stream_id, stream_control, frame, payload)

        LOGGER.debug("NsxRpcChannel::SyncCallMethodSimple polling queue: %s",
                     call.queue)
        try:
            rsp = call.queue.get(block=True, timeout=call.timeout)
        except gevent.queue.Empty:
            call.status = CallStatus(CallStatus.TIMEOUT, None)

        LOGGER.debug("NsxRpcChannel::SyncCallMethodSimple complete with status %s",
                     call.status)

        if not call.status.is_success():
            raise NsxRpcCallError(call.status)

        return rsp

    def SyncCallMethodClientStreaming(
            self,
            method_descriptor,
            # request,
            response_class,
            context):
        LOGGER.debug("NsxRpcChannel::SyncCallMethodClientStreaming")
        call_id = self.NextCallId()
        call = self.call[call_id] = self._Call(
            call_id=call_id,
            call_type=SYNCHRONOUS,
            method_descriptor=method_descriptor,
            response_class=response_class,
            timeout=context.timeout,
            callback=None,
            queue=gevent.queue.Queue(),
            local_done=False,
            remote_done=False,
            client_context=context.client_context,
            client_context_sent=False,
            status=None,
        )
        return Writer(self, call), Call(self, call)

    def SyncCallMethodServerStreaming(
            self,
            method_descriptor,
            request,
            response_class,
            context):
        LOGGER.debug("NsxRpcChannel::SyncCallMethodServerStreaming")
        call_id = self.NextCallId()
        call = self.call[call_id] = self._Call(
            call_id=call_id,
            call_type=SYNCHRONOUS,
            method_descriptor=method_descriptor,
            response_class=response_class,
            timeout=context.timeout,
            callback=None,
            queue=gevent.queue.Queue(),
            local_done=False,
            remote_done=False,
            client_context=context.client_context,
            client_context_sent=False,
            status=None,
        )
        stream_control, frame, payload = self.BuildFrameMsg(
            method_descriptor,
            call_id,
            request,
            client_context=call.client_context,
            status=None
        )
        self.connection.Send(self.stream_id, stream_control, frame, payload)
        return Call(self, call)

    def SyncCallMethodBidirectionalStreaming(
            self,
            method_descriptor,
            # request,
            response_class,
            context):
        LOGGER.debug("NsxRpcChannel::SyncCallMethodBidirectionalStreaming")
        call_id = self.NextCallId()
        call = self.call[call_id] = self._Call(
            call_id=call_id,
            call_type=SYNCHRONOUS,
            method_descriptor=method_descriptor,
            response_class=response_class,
            timeout=context.timeout,
            callback=None,
            queue=gevent.queue.Queue(),
            local_done=False,
            remote_done=False,
            client_context=context.client_context,
            client_context_sent=False,
            status=None,
        )
        return Writer(self, call), Call(self, call)

    def AsyncCallMethodSimple(
            self,
            method_descriptor,
            request,
            response_class,
            context):
        LOGGER.debug("NsxRpcChannel::AsyncCallMethodSimple")
        call_id = self.NextCallId()
        call = self.call[call_id] = self._Call(
            call_id=call_id,
            call_type=ASYNCHRONOUS,
            method_descriptor=method_descriptor,
            response_class=response_class,
            timeout=None,
            callback=context.callback,
            queue=gevent.queue.Queue(),
            local_done=False,
            remote_done=False,
            client_context=context.client_context,
            client_context_sent=False,
            status=None,
        )
        stream_control, frame, payload = self.BuildFrameMsg(
            method_descriptor,
            call_id,
            request,
            client_context=call.client_context,
            status=None
        )
        self.connection.Send(self.stream_id, stream_control, frame, payload)

    def AsyncCallMethodClientStreaming(
            self,
            method_descriptor,
            # request,
            response_class,
            context):
        LOGGER.debug("NsxRpcChannel::AsyncCallMethodClientStreaming")
        call_id = self.NextCallId()
        call = self.call[call_id] = self._Call(
            call_id=call_id,
            call_type=ASYNCHRONOUS,
            method_descriptor=method_descriptor,
            response_class=response_class,
            timeout=None,
            callback=context.callback,
            queue=gevent.queue.Queue(),
            local_done=False,
            remote_done=False,
            client_context=context.client_context,
            client_context_sent=False,
            status=None,
        )
        return Writer(self, call)

    def AsyncCallMethodServerStreaming(
            self,
            method_descriptor,
            request,
            response_class,
            context):
        LOGGER.debug("NsxRpcChannel::AsyncCallMethodServerStreaming")
        call_id = self.NextCallId()
        call = self.call[call_id] = self._Call(
            call_id=call_id,
            call_type=ASYNCHRONOUS,
            method_descriptor=method_descriptor,
            response_class=response_class,
            timeout=None,
            callback=context.callback,
            queue=gevent.queue.Queue(),
            local_done=False,
            remote_done=False,
            client_context=context.client_context,
            client_context_sent=False,
            status=None,
        )
        stream_control, frame, payload = self.BuildFrameMsg(
            method_descriptor,
            call_id,
            request,
            client_context=call.client_context,
            status=None
        )
        self.connection.Send(self.stream_id, stream_control, frame, payload)

    def AsyncCallMethodBidirectionalStreaming(
            self,
            method_descriptor,
            # request,
            response_class,
            context):
        LOGGER.debug("NsxRpcChannel::AsyncCallMethodBidirectionalStreaming")
        call_id = self.NextCallId()

        call = self.call[call_id] = self._Call(
            call_id=call_id,
            call_type=ASYNCHRONOUS,
            method_descriptor=method_descriptor,
            response_class=response_class,
            timeout=None,
            callback=context.callback,
            queue=gevent.queue.Queue(),
            local_done=False,
            remote_done=False,
            client_context=context.client_context,
            client_context_sent=False,
            status=None,
        )
        return Writer(self, call)

    def Open(self):
        LOGGER.debug("NsxRpcChannel::Open")
        stream_control = nsx_rpc_pb2.StreamControlMsg()
        stream_control.service = self.rpc_prefix
        if self.stream_destination:
            stream_control.destination = self.stream_destination
        stream_control.open_stream = True
        stream_control.priority = self.priority
        LOGGER.debug("NsxRpcChannel::Open stream %s", stream_control)
        self.connection.Send(self.stream_id, stream_control.SerializeToString(),
                             None, None)

    def Close(self, reason = CloseReason(CloseReason.UNKNOWN_REASON)):
        LOGGER.debug("NsxRpcChannel::Close %s", reason)
        if not self.is_open:
            return

        try:
            # Try sending the close messsage
            self.is_open = False
            if self.connection.is_active:
                stream_control = nsx_rpc_pb2.StreamControlMsg()
                stream_control.close_stream = True
                LOGGER.debug("NsxRpcChannel::Close stream %s", stream_control)
                self.connection.Send(self.stream_id, stream_control.SerializeToString(),
                                     None, None)
        except Exception as e:
            LOGGER.debug("NsxRpcChannel::Undelivered close stream, \
                server side may have already closed the connection.")
        finally:
            # Notify calls of connection error
            for call in list(self.call.values()):
                call.remote_done = True
                call.local_done = True
                call.status = CallStatus(CallStatus.COMMUNICATION_ERROR, "%s" % reason)
                if call.call_type is SYNCHRONOUS:
                    call.queue.put(None)
                elif call.call_type is ASYNCHRONOUS:
                    call.callback(message=None, status=call.status)
                else:
                    assert False, "unknown call_type %s" % call.call_type
            self.call.clear()
