# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential

import json
import sha.config
import sys
import time
import traceback

from sha.contrib.metric.abc_base import Metric
from sha.contrib.metric.utils import run_command

from subprocess import CalledProcessError

lr_type_map = {
   'SERVICE_ROUTER_TIER0': 'T0',
   'SERVICE_ROUTER_TIER1': 'T1',
   'VRF_SERVICE_ROUTER_TIER0': 'T0',
}

class EdgeBase(Metric):
    def __init__(self):
        self._logger = sha.config.get_logger(__name__)

    @property
    def wavefront_source(self):
        # Get hostname as source to export metrics to wavefront
        return sha.config.HOST_NAME.lower()

    def run_command_with_time_check(self, cmd, expected_sec = 1, expected_failure = False, timeout=None):
       start = time.monotonic()
       try:
           output = run_command(cmd, shell=False, timeout=timeout) or ''
       except CalledProcessError as e:
           if expected_failure:
               self._logger.debug("Failed to run command: {}, error: {}".format(e, e.stderr))
           else:
               self._logger.warning("Failed to run command: {}, error: {}".format(e, e.stderr))
           raise e
       execute_sec = time.monotonic() - start
       if execute_sec > expected_sec:
           self._logger.warning("Run '{}' for more than {} second(s): {}".format(cmd, expected_sec, execute_sec))
       return output

    def get_lr_list(self, lr_type=['SERVICE_ROUTER_TIER0']):
        # TODO : change LR list to LR dictionary. Look up can be faster
        lr_dict = {}
        # Get Tier0 LR via command
        # E.g:
        #   {
        #    "arp_refresh_timeout": 1200,
        #    "effective_uplink_mtu": 1500,
        #    "fib_v4_self": "true",
        #    "fib_v6_self": "true",
        #    "name": "SR-Tier0-LR-1",
        #    "number_arp_entries": 0,
        #    "peer_vrf": 1,
        #    "port_count": 5,
        #    "type": "SERVICE_ROUTER_TIER0",
        #    "uuid": "960c05a6-5176-43ad-aa1e-6c02a49e7fe4",
        #    "vdr": 1025,
        #    "vrf": 2
        #  },
        #
        # Cost of nsxcli is much higher than edge backend command
        #     """nsxcli -c 'get logical-routers | json'"""

        # We need 2 commands to get the mp_router_ids and the LR router names.
        mp_router_list = []
        cmd_get_mp = ['edge-appctl', '-t', '/var/run/vmware/edge/nsxa.ctl', 'lrouter/show']
        output = self.run_command_with_time_check(cmd_get_mp)
        mp_router_list = json.loads(output)
        mp_router_dict = {}
        for mp in mp_router_list:
            mp_router_dict[mp['uuid']] = mp['mp_router_id']

        cmd_get_lr = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'lrouter/show']
        output = self.run_command_with_time_check(cmd_get_lr)

        vdr_list = json.loads(output)
        for vdr in vdr_list:
            if vdr['type'] in lr_type:
                lr_dict[vdr['uuid']] = {'uuid': vdr['uuid'], 'name': vdr['name'], 'type':lr_type_map[vdr['type']], 'mp_router_id':mp_router_dict.get(vdr['uuid'],None)}

        if not lr_dict:
            self._logger.info("No {} lr found".format(lr_type))
            return None

        return lr_dict

    def get_fw_interface_list(self, lr_type = ['SERVICE_ROUTER_TIER0', 'SERVICE_ROUTER_TIER1']):
        # Get T0/T1 LR
        lr_dict = self.get_lr_list(lr_type)

        if not lr_dict:
            return None

        # Get interface. E.g:
        #  {
        #    "ICMP Active/Max": "0/10000",
        #    "Other Active/Max": "0/10000",
        #    "TCP Half Opened Active/Max": "0/1000000",
        #    "UDP Active/Max": "0/100000",
        #    "config-loaded": "true",
        #    "connection-count": "1",
        #    "iface-of": "960c05a6-5176-43ad-aa1e-6c02a49e7fe4",
        #    "iftype": "UPLINK",
        #    "ifuuid": "81ecb22d-518f-4560-a95b-7a5eaf01e758",
        #    "ike-flag": "false",
        #    "name": "plr1_uplink1",
        #    "parent-name": "SR-Tier0-LR-1",
        #    "pbr-flag": "false",
        #    "pvi": 1,
        #    "rule-flag": "true",
        #    "sync-enabled": "false",
        #    "vrf": 2
        #  },

        # The cost of nsxcli is much higher than edge backend command, comment it out
        #     """nsxcli -c 'get firewall interfaces | json'"""

        cmd_get_interface = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'fw/show']
        output = self.run_command_with_time_check(cmd_get_interface)

        interface_list = json.loads(output)
        self._logger.debug("fw interfaces: {}".format(interface_list))
        for interface in interface_list:
            if lr_dict.get(interface['iface-of'], None) is not None:
                if interface['parent-name'] == lr_dict[interface['iface-of']]['name'] and \
                   interface['iftype'] == "UPLINK":
                    lr_dict[interface['iface-of']]['interface'] = interface['ifuuid']

        self._logger.debug("get edge lr dict: {}".format(lr_dict))
        return lr_dict

    def usage(self, *args, **kwargs): pass
