# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential

import importlib
import sha.config
import sha.preconfig
import sys
import traceback

from threading import Lock

from sha.constants import PLUGIN, PLUGIN_ID, METRICS_STR, INTERVAL
from sha.config import DEDICATED_THREAD_METRICS


# Metric Registration prop:
# {
#   # UUid of the plugin
#   'plugin_id': plugin id,
#   # list of the requested data
#   # each data is located based on key of the data(name of the metric)
#   'metrics': [
#          key1,
#          key2
#   ],
#   # collect interval
#   'interval': interval
# }
#

class MetricRegister:
    """ Metric register map of collectoer
        1. plugin_metric_map:
        {
              plugin_id1: {
                  'metric': [key1, key2],
                  'interval': interval1
              },
              plugin_id2: {
                  'metric': [key2],
                  'interval': interval2
              },
        }
        2. metric_plugin_map:
        {
           key1: {
              'plugin': [
                  plugin_id1,
              ]
              'interval': interval1,
           },
           key2: {
              'plugin': [
                  plugin_id1,
                  plugin_id2,
              ]
              'interval': min(interval1, interval2)
           },
        }
    """
    def __init__(self):
        self._logger = sha.preconfig.get_sub_logger(__name__, s2comp='collector')
        self._plugin_metric_map = {}
        self._metric_plugin_map = {}
        self._plugin_metric_lock = Lock()

    def register(self, plugin_id, required_metrics, interval):
        with self._plugin_metric_lock:
            legacy_metrics = self._plugin_metric_map.get(plugin_id, None)
            if legacy_metrics:
                self._logger.warning('Plugin {} has registered before: {}'.format(plugin_id, legacy_metrics))
                raise Exception('Plugin has registed'.format(plugin_id))

            # metrics validation, skip SHA private metrics
            metrics = []
            for metric in required_metrics:
                if metric in sha.config.SHA_PRVIATE_METRICS:
                    self._logger.info('SHA private metric {metric}, no need to register.'.format(metric=metric))
                    continue
                if metric not in sha.config.ENABLED_METRICS.keys():
                    self._logger.warning('{metric} is not supported'.format(metric=metric))
                    raise Exception('Unsupported metric {}'.format(metric))
                metrics.append(metric)

            # plugin-metric mapping
            prop = {
                METRICS_STR: metrics,
                INTERVAL: interval
            }
            self._plugin_metric_map[plugin_id] = prop

            # metric-plugin mapping
            for metric in metrics:
                item = self._metric_plugin_map.get(metric, None)
                if item:
                    item[PLUGIN].append(plugin_id)
                    if interval < item[INTERVAL]:
                        item[INTERVAL] = interval
                        self._logger.info('Plugin {} is registered for metric {}, collect interval is changed to: {}'
                                          .format(plugin_id, metric, interval))
                else:
                   self._metric_plugin_map[metric] = {
                       PLUGIN: [plugin_id],
                       INTERVAL: interval
                   }

        self._logger.info('Plugin {} registered successfully with: {}, {}'.format(plugin_id, metrics, interval))
        self._logger.debug('Plugin-metric map: {}'.format(self._plugin_metric_map))
        self._logger.debug('Metric-plugin map: {}'.format(self._metric_plugin_map))


    def unregister(self, plugin_id):
        with self._plugin_metric_lock:
            prop = self._plugin_metric_map.pop(plugin_id, None)
            if not prop:
                self._logger.info('Plugin {} has not registered.'.format(plugin_id))
                return
            for metric in prop[METRICS_STR]:
                try:
                    self._metric_plugin_map[metric][PLUGIN].remove(plugin_id)
                    if not self._metric_plugin_map[metric][PLUGIN]:
                        del self._metric_plugin_map[metric]
                        self._logger.info('Plugin {} is unregistered from metric {}, no plugin listening it any more.'
                                          .format(plugin_id, metric))
                    elif prop[INTERVAL] ==  self._metric_plugin_map[metric][INTERVAL]:
                        # Re-calculate interval
                        interval = sys.maxsize
                        for plugin_id in self._metric_plugin_map[metric][PLUGIN]:
                            interval = min(interval, self._plugin_metric_map[plugin_id][INTERVAL])
                        if self._metric_plugin_map[metric][INTERVAL] != interval:
                            self._metric_plugin_map[metric][INTERVAL] = interval
                            self._logger.info('Plugin {} is unregistered from metric {}, collect interval is changed to: {}'
                                              .format(plugin_id, metric, self._metric_plugin_map[metric][INTERVAL]))
                except Exception as e:
                    trace = traceback.format_exc()
                    self._logger.warning("Failed in DataWhiteboard, Exception:{}, trace:{}".format(e, trace))

        self._logger.info('Plugin {} unregistered'.format(prop))
        self._logger.debug('Plugin-metric map: {}'.format(self._plugin_metric_map))
        self._logger.debug('Metric-plugin map: {}'.format(self._metric_plugin_map))

    def get_metric_interval(self, key):
        with self._plugin_metric_lock:
            interval = self._metric_plugin_map.get(key, {}).get(INTERVAL, None)
            return max(interval, sha.config.GLOBAL_MINIMUM_CHECKING_INTERVAL) if interval else None



class Collector:
    _metrics = dict()
    _monitors = dict()
    _logger = sha.preconfig.get_sub_logger(__name__, s2comp = 'collector')
    _metrics_module = importlib.import_module('sha.contrib.metric')
    _monitors_module = importlib.import_module('sha.contrib.monitor')
    _metric_register= MetricRegister()
    _lock = Lock()

    @classmethod
    def _load_metric(cls, metric_name):
        try:
            if metric_name not in sha.config.ENABLED_METRICS.keys():
                cls._logger.warning('{metric_name} is not allowed to load'.format(metric_name=metric_name))
                return False
            cls._metrics[metric_name] = getattr(cls._metrics_module, sha.config.ENABLED_METRICS[metric_name])()
            return True
        except AttributeError:
            cls._logger.warning('could not load {metric_name}'.format(metric_name=metric_name))
            return False

    @classmethod
    def release_metric_resource(cls, metric_name):
        """
        Release metric resources for dedicated thread metrics.
        """
        try:
            if metric_name not in DEDICATED_THREAD_METRICS:
                cls._logger.info('{} is not a dedicated thread metric'.format(metric_name))
                return

            with cls._lock:
                if metric_name in cls._metrics:
                    cls._metrics[metric_name].clean()
                    del cls._metrics[metric_name]
                    cls._logger.info('released resource for metric: {}'.format(metric_name))
        except Exception:
            trace = traceback.format_exc()
            cls._logger.warning('failed in release_metric_resource for metric:{}, trace:{}'
                                .format(metric_name, trace))

    @classmethod
    def register_metric(cls, plugin_id, metrics, interval):
        if not metrics:
            cls._logger.warning('Invalid metrics.')
            raise Exception('Invalid metrics')

        cls._metric_register.register(plugin_id, metrics, interval)

    @classmethod
    def unregister_metric(cls, plugin_id):
        cls._metric_register.unregister(plugin_id)

    @classmethod
    def get_metric(cls, metric_name):
        metric_interval = cls._metric_register.get_metric_interval(metric_name)
        if not metric_interval:
            return None, None

        with cls._lock:
            if metric_name not in cls._metrics:
                if not cls._load_metric(metric_name):
                    raise Exception('Failed to load metric {}'.format(metric_name))
            return cls._metrics[metric_name].usage, metric_interval

    @classmethod
    def _load_monitor(cls, monitor_name):
        try:
            if monitor_name not in sha.config.ENABLED_MONITORS.keys():
                cls._logger.warning('{monitor_name} is not allowed to load'.format(monitor_name=monitor_name))
                return False
            cls._monitors[monitor_name] = getattr(cls._monitors_module, monitor_name)()
            return True
        except AttributeError:
            cls._logger.warning('could not load {monitor_name}'.format(monitor_name=monitor_name))
            return False

    @classmethod
    def get_monitor(cls, monitor_name):
        if monitor_name not in cls._monitors:
            if not cls._load_monitor(monitor_name):
                return None
        return cls._monitors[monitor_name]
