# ***************************************************************************
# Copyright 2016-2021 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import errno
import grp
import os
import pwd
import re
import shutil
from subprocess import PIPE
from subprocess import Popen
import sys
import time
import traceback

sys.path.append("/opt/vmware/nsx-common/python")

import nsx_logging


# ###########################################################################
# IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT  IMPORTANT
# ###########################################################################
#
# If you are updating this file, keep in mind that this file is used by NSX
# services running on hosts (e.g. ESX, RHEL, Ubuntu) registered with the MP.
# This means that Python 3 may be used; therefore, your code has to be both
# Python 2 & 3 compatible.
#
# ###########################################################################
def invoke_command(cmd_list, stdin=None, env=None):
    """Invoke specified cmd_list with optional stdin and env

    Returns:
        (returncode, stdout, stderr) tuple where stdout and stderr values
        have been normalized to strings (i.e. account for Python3
        subprocess.communicate returning byte strings).
    """
    if sys.version_info >= (3,) and stdin is not None:
        # In Python3, communicate() expects stdin as a byte string
        stdin = bytes(stdin, "utf8")
    po = Popen(
        cmd_list, stdin=(PIPE if stdin else None), stdout=PIPE, stderr=PIPE,
        env=env)
    out, err = po.communicate(input=stdin)
    if sys.version_info >= (3,):
        # In Python3, communicate() returns out and err as byte strings
        if out is not None:
            out = out.decode("utf8")
        if err is not None:
            err = err.decode("utf8")
    return po.returncode, out, err


_BUNDLE_TIMEOUT = 5400  # Abort bundle collection via API after this num seconds
_NAPI_DIR = "/opt/vmware/nsx-node-api/bin/python"
_NSX_PYTHON_COMMON = "/opt/vmware/nsx-common/python"
_PYTHONPATH = _NAPI_DIR + os.path.pathsep + _NSX_PYTHON_COMMON
_NAPI_SUPPORT_BUNDLE_SCRIPT = (
    "%s/management_api/napi/root/administration/support_bundles/__self__.py" %
    _NAPI_DIR)
if not os.path.exists(_NAPI_SUPPORT_BUNDLE_SCRIPT):
    _NAPI_SUPPORT_BUNDLE_SCRIPT = (
        "%s/management_api/napi/root/administration/support_bundles/__self__.pyo" %
        _NAPI_DIR)
_KVM_BUNDLE_SCRIPT = "/opt/vmware/nsx-agent/bin/support_save.py"
_ESX_BUNDLE_SCRIPT = "/bin/vm-support"

_SBUNDLE_USR = "nsx-opsagent"
_SBUNDLE_GRP = "nsx"

_NODETYPE_CENTOSCONTAINER = "nsx-centoscontainer"
_NODETYPE_CENTOSKVM = "nsx-centos"
_NODETYPE_CENTOSSERVER = "nsx-centosserver"
_NODETYPE_CONTROL = "nsx-controller"
_NODETYPE_EDGE = "nsx-edge"
_NODETYPE_ESXI = "nsx-esxi"
_NODETYPE_GM = "global-manager"
_NODETYPE_HOST = "nsx-host"
_NODETYPE_KEYMANAGER = "nsx-keymanager"
_NODETYPE_MANAGER = "nsx-manager"
_NODETYPE_PUBLIC_CLOUD_GATEWAY = "nsx-public-cloud-gateway"
_NODETYPE_RHELCONTAINER = "nsx-rhelcontainer"
_NODETYPE_RHELKVM = "nsx-rhelkvm"
_NODETYPE_RHELSERVER = "nsx-rhelserver"
_NODETYPE_UBUNTUKVM = "nsx-ubuntukvm"
_NODETYPE_UBUNTUSERVER = "nsx-ubuntuserver"
_NODETYPE_SLESKVM = "nsx-sleskvm"
_NODETYPE_SLESSERVER = "nsx-slesserver"
_NODETYPE_OELSERVER = "nsx-oelserver"
_NODETYPE_UNKNOWN = "Unknown"

_OK = 0
_ERROR_UNKNOWN_ERROR = 1
_ERROR_UNKNOWN_NODETYPE = 2
_ERROR_NOT_ENOUGH_ARGUMENTS_SPECIFIED = 3

_TIMEOUT_BIN = "/usr/bin/timeout"
_TIMEOUT_ERROR_EXIT = 124
_TIMEOUT_ERROR_KILL = 137
_TIMEOUT_ERROR_UNKNOWN = -9
_TIMEOUT_ERRORS = [
    _TIMEOUT_ERROR_EXIT, _TIMEOUT_ERROR_KILL, _TIMEOUT_ERROR_UNKNOWN]

_FILE_REDHAT_RELEASE = "/etc/redhat-release"
_FILE_OEL_RELEASE = "/etc/oracle-release"


def _get_node_type():
    result = _NODETYPE_UNKNOWN
    try:
        with open("/etc/nsx_issue") as fo:
            lines = fo.read().split("\n")
        for line in lines:
            parts = line.strip().split()
            if parts and parts[0] == "node-type:":
                # Unified appliances can report multiple node-type values in
                # /etc/nsx_issue, if any of the values is nsx-manager, return
                # nsx-manager so that if new unified appliance types get added
                # in the future and is really a manager, just report nsx-manager
                if _NODETYPE_MANAGER in parts[1:]:
                    return _NODETYPE_MANAGER
                return parts[-1]
    except Exception:
        lg.warning(
            "Unable to determine node type\n%s\n" %
            traceback.format_exc())
        sys.stderr.write(
            "Unable to determine node type\n%s\n" %
            traceback.format_exc())
    return result

def _get_os_type():
    cmd = ["/usr/bin/lsb_release", "-si"]
    try:
        result, out, err = invoke_command(cmd);
        return out
    except Exception:
        return None


def main(argv):
    # Verify destination directory specified in argv
    if len(argv) < 5:
        lg.warning("Not enough arguments specified")
        sys.stderr.write("Not enough arguments specified")
        return _ERROR_NOT_ENOUGH_ARGUMENTS_SPECIFIED, None
    result = _OK
    dst_dir = argv[1]
    node_id = argv[2]
    content_filter = argv[3]
    log_age_limit = argv[4]
    dst_bundle_path = None
    all_files = content_filter == "ALL"
    # Invoke command(s) to create the bundle, note the command(s) may be
    # different depending on the current node type
    node_type = _get_node_type()
    if node_type == "nsx-esx":
        uid = 0
        gid = 0
    else:
        uid = pwd.getpwnam(_SBUNDLE_USR).pw_uid
        gid = grp.getgrnam(_SBUNDLE_GRP).gr_gid
    if node_type in [
            _NODETYPE_CONTROL, _NODETYPE_EDGE, _NODETYPE_KEYMANAGER,
            _NODETYPE_MANAGER, _NODETYPE_PUBLIC_CLOUD_GATEWAY, _NODETYPE_GM]:
        cmd = [
            _TIMEOUT_BIN, "--kill-after=1", str(_BUNDLE_TIMEOUT),
            "/usr/bin/python3", "-OO", _NAPI_SUPPORT_BUNDLE_SCRIPT,
            dst_dir, str(uid), str(gid), str(all_files), log_age_limit]
        try:
            env = {"PYTHONPATH": _PYTHONPATH}
            result, out, err = invoke_command(cmd, env=env)
            if result == 0:
                out = out.strip()
                tmp_path = os.path.join(dst_dir, os.path.basename(out))
                shutil.move(out, tmp_path)
                dst_bundle_path = tmp_path
            else:
                # Map timeout result from /usr/bin/timeout to 124. It can be:
                # * 124 or 137 as discussed in man timeout(1)
                # * -9 due to exit via signal (presumably when /usr/bin/timeout
                #   kills python -OO _NAPI_SUPPORT_BUNDLE_SCRIPT ...)
                if result in _TIMEOUT_ERRORS:
                    result = _TIMEOUT_ERROR_EXIT
                    lg.warning("Support bundle generation timed out")
                    sys.stderr.write("Support bundle generation timed out")
                else:
                    lg.warning(
                        "Unable to collect support bundle: result: %s, "
                        "out: %s, err: %s" % (result, out, err))
                    sys.stderr.write(
                        "Unable to collect support bundle: result: %s, "
                        "out: %s, err: %s" % (result, out, err))
        except Exception:
            result = _ERROR_UNKNOWN_ERROR
            lg.warning(
                "Failed to generate or copy support bundle to "
                "destination directory\n%s\n" % traceback.format_exc())
            sys.stderr.write(
                "Failed to generate or copy support bundle to "
                "destination directory\n%s\n" % traceback.format_exc())
    elif node_type in ["nsx-kvm", "nsx-bms", "nsx-bmc"]:
        # _KVM_BUNDLE_SCRIPT supports specifying a target path or letting
        # it choose where to generate the target. Unfortunately, letting
        # it choose means we have to parse stdout which is brittle, so we
        # specify the target path.
        ntype = _NODETYPE_HOST
        env = {}
        if os.path.exists("/etc/lsb-release"): #UBUNTU
            if node_type == "nsx-kvm":
                ntype = _NODETYPE_UBUNTUKVM
            elif node_type == "nsx-bms":
                ntype = _NODETYPE_UBUNTUSERVER
            env["PATH"] = (
                "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin")
        elif os.path.exists("/etc/SuSE-release"): #SUSE
            if node_type == "nsx-kvm":
                ntype = _NODETYPE_SLESKVM
            elif node_type == "nsx-bms":
                ntype = _NODETYPE_SLESSERVER
            env["PATH"] = "/sbin:/usr/sbin:/usr/local/sbin:/root/bin:/usr/local/bin:/usr/bin:/bin"
        elif os.path.exists(_FILE_REDHAT_RELEASE): #RHEL
            os_type = _get_os_type()
            if node_type == "nsx-kvm":
                redhat_release_contents = ""
                try:
                    with open(_FILE_REDHAT_RELEASE) as fo:
                        redhat_release_contents = fo.read()
                except Exception:
                    pass
                ntype = (
                    _NODETYPE_CENTOSKVM if "CentOS" in redhat_release_contents
                    else _NODETYPE_RHELKVM)
            elif node_type == "nsx-bms":
                if os_type == "RedHatEnterpriseServer" or os_type == "RedHatEnterprise":
                    ntype = _NODETYPE_RHELSERVER
                elif os_type == "CentOS":
                    ntype = _NODETYPE_CENTOSSERVER
            elif node_type == "nsx-bmc":
                if os_type == "RedHatEnterpriseServer":
                    ntype = _NODETYPE_RHELCONTAINER
                elif os_type == "CentOS":
                    ntype = _NODETYPE_CENTOSCONTAINER
            env["PATH"] = "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin"
        elif os.path.exists(_FILE_OEL_RELEASE): #OEL
            if node_type == "nsx-bms":
                ntype = _NODETYPE_OELSERVER
            env["PATH"] = "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin"
        bundle_name = "%s_%s_%s.tgz" % (
            ntype.replace("-", "_"), node_id,
            time.strftime("%Y%m%d_%H%M%S"))
        bundle_path = os.path.join(dst_dir, bundle_name)
        try:
            os.unlink(bundle_path)
        except OSError as ex:
            if ex.errno != errno.ENOENT:
                lg.warning(
                    "Failed to cleanup old bundle files: %s" %
                    traceback.format_exc())
                sys.stderr.write(
                    "Failed to cleanup old bundle files: %s" %
                    traceback.format_exc())
        try:
            # If timeout command exists, limit total time to _BUNDLE_TIMEOUT
            # like we do on appliances
            cmd = []
            timeout_configured = False
            if os.path.isfile(_TIMEOUT_BIN) and os.access(_TIMEOUT_BIN, os.X_OK):
                cmd.extend(
                    [_TIMEOUT_BIN, "--kill-after=1", str(_BUNDLE_TIMEOUT)])
                timeout_configured = True
            cmd.extend([sys.executable, _KVM_BUNDLE_SCRIPT, "-t", bundle_path, "-f",
                   content_filter.upper()])
            if log_age_limit != "0":
                cmd.extend(["-l", log_age_limit])
            result, out, err = invoke_command(cmd, env=env)
            if result == 0:
                dst_bundle_path = bundle_path
            elif timeout_configured and result in _TIMEOUT_ERRORS:
                result = _TIMEOUT_ERROR_EXIT
                lg.warning("Support bundle generation timed out")
                sys.stderr.write("Support bundle generation timed out")
            elif err:
                lg.warning(
                    "Unable to collect support bundle, out: %s, err: %s"
                    % (out, err))
                sys.stderr.write(
                    "Unable to collect support bundle, out: %s, err: %s"
                    % (out, err))
        except Exception:
            result = _ERROR_UNKNOWN_ERROR
            lg.warning(
                "Failed to generate or copy support bundle to destination "
                "directory\n%s\n" % traceback.format_exc())
            sys.stderr.write(
                "Failed to generate or copy support bundle to destination "
                "directory\n%s\n" % traceback.format_exc())
    elif node_type == "nsx-esx":
        # dst_dir is not used on ESXi
        bundle_name = "%s_%s_%s.tgz" % (
            _NODETYPE_ESXI.replace("-", "_"), node_id,
            time.strftime("%Y%m%d_%H%M%S"))
        try:
            env = {"LANG": "en_US.UTF-8", "LC_ALL": "en_US.UTF-8"}
            result, out, err = invoke_command([_ESX_BUNDLE_SCRIPT], env=env)
            if result == 0:
                # The ESX support bundle generation script outputs the following
                # to stdout, we extract the location of the generated bundle
                # file using regex:
                #
                # Please attach this file when submitting an incident report.
                # To file a support incident, go to http://www.vmware.com/...
                # To see the files collected, check '/var/tmp/FILENAME.tgz'
                match = re.search("^.* check '(.+)'.*$", out, re.MULTILINE)
                if match is not None:
                    # Create a symlink using the standard NSX bundle name format,
                    # the caller of this script is responsible for deleting the
                    # symlink and bundle file
                    output_path = match.group(1)
                    try:
                        link_dir = os.path.realpath(os.path.dirname(output_path))
                    except UnicodeEncodeError:
                        link_dir = os.path.realpath(
                            os.path.dirname(
                                output_path).encode("utf-8")).decode("utf-8")
                    bundle_path = os.path.join(link_dir, bundle_name)
                    try:
                        try:
                            os.symlink(output_path, bundle_path)
                        except UnicodeEncodeError:
                            os.symlink(
                                output_path.encode("utf-8"),
                                bundle_path.encode("utf-8"))
                    except OSError as ex:
                        # If ESXi is configured to use VSAN, symlinks may not
                        # be supported on the filesystem where the support
                        # bundle has been output. So we create the temp symlink
                        # in /tmp instead.
                        if hasattr(ex, "errno") and ex.errno == errno.ENOSYS:
                            bundle_path = os.path.join("/tmp", bundle_name)
                            try:
                                os.symlink(output_path, bundle_path)
                            except UnicodeEncodeError:
                                os.symlink(
                                    output_path.encode("utf-8"),
                                    bundle_path.encode("utf-8"))
                        else:
                            raise
                    dst_bundle_path = bundle_path
                else:
                    result = _ERROR_UNKNOWN_ERROR
                    lg.warning(
                        "Unexpected output from vm-support: %s" % out)
                    sys.stderr.write(
                        "Unexpected output from vm-support: %s" % out)
            elif err:
                lg.warning(
                    "Unable to collect support bundle: rc: %s, error: %s"
                    % (result, err))
                sys.stderr.write(
                    "Unable to collect support bundle: rc: %s, error: %s"
                    % (result, err))
        except Exception:
            result = _ERROR_UNKNOWN_ERROR
            lg.warning(
                "Failed to generate or copy support bundle to destination "
                "directory\n%s\n" % traceback.format_exc())
            sys.stderr.write(
                "Failed to generate or copy support bundle to destination "
                "directory\n%s\n" % traceback.format_exc())
    else:
        result = _ERROR_UNKNOWN_NODETYPE
    return result, dst_bundle_path


if __name__ == "__main__":
    global lg
    nsx_logging.basicConfig(syslog=True, subcomp="sbundle-collect")
    lg = nsx_logging.getLogger(__name__)
    lg.setLevel(nsx_logging.INFO)
    result, dst_bundle_path = main(sys.argv)
    if result == _OK:
        sys.stdout.write(dst_bundle_path)
    sys.exit(result)
